//
// counted_ptr.hh: counted reference wrapper
// Copyright (C) 2004 Benoit Perrot <benoit@lrde.epita.fr>
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// It is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//

/** \file misc/counted_ptr.hh
    \brief Counted reference wrapper.

    A great part of the implementation comes from LRDE's ref.hh/ref.hxx,
    provided in Tiger Compiler by Raphael Poss. */

#ifndef MISC_COUNTED_PTR_HH
# define MISC_COUNTED_PTR_HH

# include "misc/contract.hh"

namespace misc
{

  template < typename T >
  /** \brief A reference counted smart pointer which provides a limited form
      of garbage collection.

      This class implements the proposal of Gregory Colvin in Exception
      Safe Smart Pointers, C++ committee document 94-168/N0555, July, 1994.

      The concept of counted pointers is very similar to the concept of
      file links in file systems, that is why I choose to use the
      well-known file manipulation command names as method identifiers.
  */
  class counted_ptr
  {
  public:
    /// The pointed-to type.
    typedef T element_type;

  private:
    counted_ptr(element_type *ptr, long *count):
      ptr_(ptr), count_(count)
    {
      ++(*count_);
    }

    template < typename U >
    friend class counted_ptr;

    /** \name Constructor and destructor
        \{ */
  public:
    /** \brief A counted_ptr is usually constructed from a raw pointer.
        \param ptr A pointer (defaults to NULL). */
    counted_ptr(element_type *ptr = 0):
      ptr_(ptr)
    {
      count_ = new long(1);
    }

    /** \brief A counted_ptr can be constructed from another counted_ptr.
        \param c Another %auto_ptr of the same type. */
    counted_ptr(const counted_ptr &c):
      ptr_(c.ptr_), count_(c.count_)
    {
      ++(*count_);
    }

    /** \brief Release a link to ptr.
     **
     ** When a counted_ptr is deleted, its counter is decremented; if
     ** it reaches zero, the pointed object is deleted. */
    ~counted_ptr()
    {
      unlink();
    }
    /** \} */

    /** \name Conversion, assignement, comparison
        \{ */
  public:
    /** \brief counted_ptr conversion operator.

        This conversion operator avoids having to define several
        constructors and test operators for creation from or comparison with
        counted pointers of a different but related type.
    */
    template < typename U >
    operator counted_ptr< U >() const
    {
      return counted_ptr< U >(ptr_, count_);
    }

    /** \brief counted_ptr assignment operator.
        \param c Another counted_ptr of the same type. */
    counted_ptr&
    operator=(const counted_ptr &c)
    {
      link(c.ptr_, c.count_);
      return *this;
    }

    /** \brief Return true if \a c and this share the same data. */
    bool
    operator==(const counted_ptr &c)
    {
      return this->count_ == c.count_;
    }
    /** \brief Return true if \a c and this do not share the same data. */
    bool
    operator!=(const counted_ptr &c)
    {
      return this->count_ != c.count_;
    }
    /** \} */

  private:
    void unlink()
    {
      --(*count_);
      if (!*count_)
        {
          delete ptr_;
          delete count_;
        }
    }

    void link(T *ptr, long *count)
    {
      if (count_ != count)
        {
          ++(*count);
          unlink();
          ptr_ = ptr;
          count_ = count;
        }
    }

    /** \name Dereferencing
        \{ */
  public:
    bool
    exists() const
    {
      return ptr_ != 0;
    }
    operator bool() const
    {
      return ptr_ != 0;
    }

    element_type&
    operator*() const
    {
      precondition(ptr_ != 0);
      return *ptr_;
    }

    element_type*
    operator->() const
    {
      precondition(ptr_ != 0);
      return ptr_;
    }
    /** \} */


  private:
    /// Contained pointer
    T *ptr_;
    /// Reference counter
    long *count_;
  };

} // namespace misc

#endif // !MISC_COUNTED_PTR_HH
